#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include "iolib.h"
#define PROGNAME "ratelimit"
#include "freedt.h"

/* The size of the data buffers either way. */
#define BUFFERSIZE 64

pid_t pid;

void recordio(char *data, int length, char *prefix) {
	if (length == 0) {
		iformat(iberr, 'i', pid, 'c', prefix, 'c', "[EOF]\n", 0);
	} else {
		do {
			char c;
	
			iformat(iberr, 'i', pid, 'c', prefix, 0);
			while (length--) {
				c = *data++;
				if (c == '\n') break;
				iwritea(iberr, &c, 1);
			}
			if (c == '\n')
				iformat(iberr, 'c', " \n", 0);
			else
				iformat(iberr, 'c', "+\n", 0);
		} while (length > 0);
	}
	iflush(iberr);
}

int main(int argc, char **argv) {
	int inlimit = 0, outlimit = 0, record = 0;
	int incount, outcount, inopen = 1, rc;
	int inpipe[2], outpipe[2];
	char inbuf[BUFFERSIZE], outbuf[BUFFERSIZE];
	int inbsize = 0, outbsize = 0;
	time_t instart = -1, outstart = -1;
	io *ifrom, *ito;

	istdinit();
	pid = getpid();

	while (1) {
		char c = getopt(argc, argv, "+i:o:r");
		if (c == -1)
			break;
		switch (c) {
		case 'i':
			inlimit = atoi(optarg);
			break;
		case 'o':
			outlimit = atoi(optarg);
			break;
		case 'r':
			record = 1;
			break;
		default:
			die("unknown option");
		}
	}

	if ((argc - optind) < 1)
		die("usage: ratelimit [-i in] [-o out] [-r] command ...");

	if (pipe(inpipe) < 0 || pipe(outpipe) < 0)
		die("unable to create pipe");

	rc = fork();
	if (rc < 0)
		die("fork failed");
	if (rc > 0) {
		close(inpipe[1]);
		close(outpipe[0]);
		dup2(inpipe[0], 0);
		dup2(outpipe[1], 1);

		execvp(argv[optind], &argv[optind]);
		die2(argv[optind], "unable to exec");
	}
	close(inpipe[0]);
	close(outpipe[1]);

	ifrom = ifromfd(outpipe[0]);
	if (ifrom == NULL)
		die("unable to create io");
	ito = ifromfd(inpipe[1]);
	if (ito == NULL)
		die("unable to create io");

	incount = outcount = 0;
	while (1) {
		int rc, maxfd = -1;
		fd_set infds, outfds;
		struct timeval tv;
		time_t now = time(NULL);

		tv.tv_sec = 1;
		tv.tv_usec = 0;

#define ADD(fd, set) \
	do { \
		FD_SET(fd, set); \
		if (fd > maxfd) \
			maxfd = fd; \
	} while (0)

		FD_ZERO(&infds);
		if (inbsize == 0)
			ADD(0, &infds);
		if (outbsize == 0)
			ADD(outpipe[0], &infds);

/* This is somewhat messy, but it's less messy than inlining it twice.
   A write is allowed if some data is available to be written,
   and some time has elapsed so that the rate can be measured,
   and the average rate so far doesn't exceed the limit. */
#define CAN_WRITE(bsize, limit, count, start) \
	((bsize > 0) && \
		((limit == 0) \
		 || ((now != start) && (count / (now - start) <= limit))))

		FD_ZERO(&outfds);
		if (inopen && CAN_WRITE(inbsize, inlimit, incount, instart))
			ADD(inpipe[1], &outfds);
		if (CAN_WRITE(outbsize, outlimit, outcount, outstart))
			ADD(1, &outfds);

		rc = select(maxfd + 1, &infds, &outfds, NULL, &tv);
		if (rc < 0)
			die("select failed");

		if (rc == 0)
			continue;

		if (FD_ISSET(0, &infds)) {
			inbsize = iread(iin, inbuf, BUFFERSIZE);
			if (inbsize < 0)
				die("read from stdin failed");
			if (record)
				recordio(inbuf, inbsize, " < ");
			if (inbsize == 0) {
				close(inpipe[1]);
				inopen = 0;
			}
			incount += inbsize;
			if (instart == -1)
				instart = time(NULL);
		}

		if (FD_ISSET(outpipe[0], &infds)) {
			outbsize = iread(ifrom, outbuf, BUFFERSIZE);
			if (outbsize < 0)
				die("read from child failed");
			if (record)
				recordio(outbuf, outbsize, " > ");
			if (outbsize == 0) {
				break;
			}
			outcount += outbsize;
			if (outstart == -1)
				outstart = time(NULL);
		}

		if (FD_ISSET(inpipe[1], &outfds)) {
			if (iwritea(ito, inbuf, inbsize) < 0)
				die("write to child failed");
			inbsize = 0;
		}

		if (FD_ISSET(1, &outfds)) {
			if (iwritea(iout, outbuf, outbsize) < 0)
				die("write to stdout failed");
			outbsize = 0;
		}
	}

	return 0;
}

